/* Implementation of polynom operations
 * Author:  Yotam Medini  yotam.medini@gmail.com -- Created: 2010/December/20 
 */

#include "polynom.h"
#include <stdlib.h>
#include <math.h>


const double polynom_epsilon = 1e-6;


polynom_t  *polynom_create(int degree)
{
    int  k;
    polynom_t  *p = malloc(sizeof(polynom_t));
    p->degree = degree;
    p->coefs = malloc((degree + 1)*sizeof(double));
    for (k = 0; k < degree; k++) { p->coefs[k] = 0; }
    p->coefs[degree] = 1;
    return p;
}


polynom_t  *polynom_create_constant(double c0)
{
    polynom_t  *p = polynom_create(0);
    p->coefs[0] = c0;
    return p;
}


void polynom_copy(polynom_t *target, polynom_t *source)
{
    int  i, d = source->degree;
    target->degree = d;
    target->coefs = realloc(target->coefs, (d + 1) * sizeof(double));
    for (i = 0;  i <= d;  i++)
    {
        target->coefs[i] = source->coefs[i];
    }
}


void polynom_destroy(polynom_t *poly)
{
    free(poly->coefs);
    free(poly);
}


double polynom_coef(polynom_t *p, int i)
{
    return (i <= p->degree ? p->coefs[i] : 0.);
}


int polynom_equal(polynom_t *p0, polynom_t *p1)
{
    int       zero = 0;
    polynom_t *d = polynom_create(0);
    polynom_subtract(d, p0, p1);
    zero = (d->degree == 0) && d->coefs[0] == 0.;
    polynom_destroy(d);
    return zero;
}


double  polynom_evaluate(polynom_t *p, double x)
{
    double y = 0;
    int  k;
    for (k = p->degree;  k >= 0;  k--)
    {
        y *= x;
        y += p->coefs[k];
    }
    return y;
}


void polynom_print_coefs(FILE *f, polynom_t *p)
{
    int  k;

    for (k = p->degree;  k >= 0;  k--)
    {
        fprintf(f, "%g ", p->coefs[k]);
    }
}


void polynom_print(FILE *f, polynom_t *p)
{
    int  k = p->degree;

    fprintf(f, "%gx^%d", p->coefs[k], k);
    for (k--;  k >= 0;  k--)
    {
        double  a = p->coefs[k];
        char    *sym = "+";
        if (a < 0)
        {
            a = -a;
            sym = "-";
        }
        fprintf(f, " %s %gx^%d", sym, a, k);
    }
}


void polynom_reduce(polynom_t *p)
{
    int  k;

    /* Zero tiny coefficients */
    for (k = p->degree;  k >= 0;  k--)
    {
        if (fabs(p->coefs[k]) < polynom_epsilon)
        {
             p->coefs[k] = 0.;
        }
    }

    /* Reduce degree */ 
    for (k = p->degree;  k > 0 && fabs(p->coefs[k]) < polynom_epsilon;  k--);
    p->degree = k;
}


void polynom_add(polynom_t *result, polynom_t *p, polynom_t *q)
{
    int  i, dmax = (p->degree < q->degree ? q->degree : p->degree);
    result->coefs = realloc(result->coefs, (dmax + 1) * sizeof(double));

    for (i = 0;  i <= dmax;  i++)
    {
        result->coefs[i] = polynom_coef(p, i) + polynom_coef(q, i);
    }
    result->degree = dmax;
    polynom_reduce(result);
}


void polynom_subtract(polynom_t *result, polynom_t *p, polynom_t *q)
{
    polynom_t  *minus1 = polynom_create_constant(-1);
    polynom_t  *minus_q = polynom_create(0);
    polynom_multiply(minus_q, minus1, q);

    polynom_add(result, p, minus_q);

    polynom_destroy(minus1);
    polynom_destroy(minus_q);
}


void polynom_multiply(polynom_t *result, polynom_t *p, polynom_t *q)
{
    int      k, d = p->degree + q->degree;
    double  *new_coefs = malloc((d + 1)*sizeof(double));

    for (k = 0;  k <= d;  k++)
    {
        double  ck = 0;
        int     j;
        for (j = 0;  j <= k;  j++)
        {
            ck += polynom_coef(p, j) * polynom_coef(q, k - j);
        }
        new_coefs[k] = ck;
    }
    result->degree = d;
    free(result->coefs);
    result->coefs = new_coefs;
}


void polynom_divide(polynom_t *q, polynom_t *r, polynom_t *n, polynom_t *d)
{
    polynom_t *t = polynom_create_constant(0.);
    polynom_copy(q, t);
    polynom_copy(r, n);
    polynom_destroy(t);

    while ((r->degree > 0) && (r->degree >= d->degree))
    {
        int tdeg = r->degree - d->degree;
        t = polynom_create(tdeg);
        t->coefs[tdeg] = r->coefs[r->degree] / d->coefs[d->degree];
        polynom_add(q, q, t);
        polynom_multiply(t, t, d);
        polynom_subtract(r, r, t);
        polynom_destroy(t);
    }
}


int polynom_is_divider(polynom_t *d, polynom_t *n)
{
    int        is_divider;
    polynom_t  *q = polynom_create(0);
    polynom_t  *r = polynom_create(0);
    polynom_divide(q, r, n, d);
    is_divider = (r->degree == 0) && (fabs(r->coefs[0]) < polynom_epsilon);
    polynom_destroy(q);
    polynom_destroy(r);
    return is_divider;
}


void polynom_gcd(polynom_t *gcd, polynom_t *p, polynom_t *q)
{
    double     high_coef;
    /* HomeWork : Euclid algorithm */

    /* Divide the gcd by constant so leading coefficient will be 1 */
    high_coef = gcd->coefs[gcd->degree];
    if (fabs(high_coef) > polynom_epsilon)
    {
        polynom_t  *c = polynom_create_constant(1./high_coef);
        polynom_multiply(gcd, gcd, c);
        polynom_destroy(c);
    }
}


void polynom_derivative(polynom_t *result, polynom_t *p)
{
    /* HomeWork */
}


int polynom_solve(double *solutions, polynom_t *p)
{
    /* Long term project */
    return 0;
}
